from django.db import models

from djeneralize.manager import SpecializationManager
from djeneralize.models import BaseGeneralizationModel

__all__ = [
    'FruitManager', 'SpecializedFruitManager', 'Fruit', 'Apple', 'Banana'
    ]

class FruitManager(models.Manager):
    """By default we don't want to return any rotten fruit"""

    def get_queryset(self):
        return super(FruitManager, self).get_queryset().filter(rotten=False)

class SpecializedFruitManager(SpecializationManager):
    """
    And any specializations also shouldn't return any rotten fruit
    specializations.

    """

    def get_queryset(self):
        return super(SpecializedFruitManager, self).get_queryset().filter(
            rotten=False
            )


class Fruit(BaseGeneralizationModel):
    """A piece of fruit"""

    name = models.CharField(max_length=30)
    rotten = models.BooleanField(default=False)

    objects = FruitManager()
    specializations = SpecializedFruitManager()

    def __unicode__(self):
        return self.name


class Apple(Fruit):
    """An apple"""

    radius = models.IntegerField()

    class Meta:
        specialization = 'apple'


class Banana(Fruit):
    """A banana"""

    curvature = models.DecimalField(max_digits=3, decimal_places=2)

    class Meta:
        specialization = 'banana'


##############################################################################
#
# Copyright (c) 2011,2013, 2degrees Limited <2degrees-floss@googlegroups.com>.
# All Rights Reserved.
#
# This file is part of djeneralize <https://github.com/2degrees/djeneralize>,
# which is subject to the provisions of the BSD at
# <http://dev.2degreesnetwork.com/p/2degrees-license.html>. A copy of the
# license should accompany this distribution. THIS SOFTWARE IS PROVIDED "AS IS"
# AND ANY AND ALL EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST
# INFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
#
##############################################################################
from tests.test_djeneralize.producers.models import EcoProducer
from tests.test_djeneralize.fruit.models import Banana

"""Fixtures for djeneralize tests"""

from decimal import Decimal as D

from fixture import DataSet


__all__ = [
    'PenData', 'FountainPenData', 'BallPointPenData', 'PencilData',
    'EcoProducerData', 'ShopData'
    ]


class PenData(DataSet):

    class Meta:
        django_model = 'writing.Pen'

    class GeneralPen:
        specialization_type = '/pen/'
        name = 'General pen'
        length = 15
        ink_colour = 'Blue'


class FountainPenData(DataSet):

    class Meta:
        django_model = 'writing.FountainPen'

    class MontBlanc:
        specialization_type = '/pen/fountain_pen/'
        name = 'Mont Blanc'
        length = 18
        ink_colour = 'Black'
        nib_width = D('1.25')

    class Parker:
        specialization_type = '/pen/fountain_pen/'
        name = 'Parker'
        length = 14
        ink_colour = 'Blue'
        nib_width = D('0.75')

class BallPointPenData(DataSet):

    class Meta:
        django_model = 'writing.BallPointPen'

    class Bic:
        specialization_type = '/pen/ballpoint_pen/'
        name = 'Bic'
        length = 12
        ink_colour = 'Blue'
        replaceable_insert = False

    class Papermate:
        specialization_type = '/pen/ballpoint_pen/'
        name = 'Papermate'
        length = 13
        ink_colour = 'Green'
        replaceable_insert = True


class PencilData(DataSet):

    class Meta:
        django_model = 'writing.Pencil'

    class Crayola:
        specialization_type = '/pencil/'
        name = 'Crayola'
        length = 8
        lead = 'B2'

    class Technical:
        specialization_type = '/pencil/'
        name = 'Technical'
        length = 12
        lead = 'H5'


class BananaData(DataSet):

    class Meta:
        django_model = 'fruit.Banana'

    class Banana:
        specialization_type = Banana.model_specialization
        name = 'Banana from Canary Islands'
        curvature = D('1.10')


class EcoProducerData(DataSet):

    class Meta:
        django_model = 'producers.EcoProducer'

    class BananaProducer:
        specialization_type = EcoProducer.model_specialization
        name = 'Ecological Producer'
        produce = BananaData.Banana
        pen = PenData.GeneralPen
        fertilizer = 'Love'


class ShopData(DataSet):

    class Meta:
        django_model = 'producers.Shop'

    class EcoMart:
        name = 'EcoMart'
        producer = EcoProducerData.BananaProducer



##############################################################################
#
# Copyright (c) 2011,2013, 2degrees Limited <2degrees-floss@googlegroups.com>.
# All Rights Reserved.
#
# This file is part of djeneralize <https://github.com/2degrees/djeneralize>,
# which is subject to the provisions of the BSD at
# <http://dev.2degreesnetwork.com/p/2degrees-license.html>. A copy of the
# license should accompany this distribution. THIS SOFTWARE IS PROVIDED "AS IS"
# AND ANY AND ALL EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST
# INFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
#
##############################################################################
import os
# Ensure that Django knows where the project is:
os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.test_djeneralize.settings'

from fixture.django_testcase import FixtureTestCase
from nose.tools import eq_

from tests.fixtures import BananaData, EcoProducerData, PenData, ShopData
from tests.test_djeneralize.fruit.models import Banana
from tests.test_djeneralize.producers.models import EcoProducer
from tests.test_djeneralize.writing.models import WritingImplement


class TestForeignKey(FixtureTestCase):

    datasets = [EcoProducerData, BananaData, PenData, ShopData]

    def test_specialized_foreign_key(self):
        """A SpecializedForeignKey field return the specialized counterpart"""
        eco = EcoProducer.objects.get(name=EcoProducerData.BananaProducer.name)
        eq_(eco.produce.__class__, Banana)
        eq_(eco.pen.__class__, WritingImplement)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2011,2013, 2degrees Limited <2degrees-floss@googlegroups.com>.
# All Rights Reserved.
#
# This file is part of djeneralize <https://github.com/2degrees/djeneralize>,
# which is subject to the provisions of the BSD at
# <http://dev.2degreesnetwork.com/p/2degrees-license.html>. A copy of the
# license should accompany this distribution. THIS SOFTWARE IS PROVIDED "AS IS"
# AND ANY AND ALL EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST
# INFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""Integration tests to confirm nothing Django-related is broken"""

import os
# Ensure that Django knows where the project is:
os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.test_djeneralize.settings'

from .fixtures import *
from .test_djeneralize.writing.models import *
from djeneralize.utils import *

from django.db.models import Avg, F, Q
from django.db.models.query import ValuesListQuerySet, ValuesQuerySet
from fixture.django_testcase import FixtureTestCase
from nose.tools import assert_not_equal
from nose.tools import eq_
from nose.tools import ok_


def compare_generalization_to_specialization(generalization, specialization):
    eq_(generalization.pk, specialization.pk)
    eq_(generalization.name, specialization.name)
    eq_(generalization.length, specialization.length)
    assert_not_equal(generalization, specialization)


class TestManager(FixtureTestCase):
    datasets = [PenData, PencilData, FountainPenData, BallPointPenData]


class TestSpecializedQuerySet(FixtureTestCase):

    datasets = [PenData, PencilData, FountainPenData, BallPointPenData]

    def _check_attributes(self, normal_objects, specialized_objects):
        """
        Helper test to run through the two querysets and
        test various attributes

        """

        for normal_object, specialized_object in zip(
            normal_objects, specialized_objects
            ):

            eq_(normal_object.__class__, WritingImplement)
            assert_not_equal(specialized_object.__class__, WritingImplement)

            compare_generalization_to_specialization(
                normal_object,
                specialized_object
                )

            ok_(isinstance(specialized_object, WritingImplement))

    def test_all(self):
        """Check the all() method works correctly"""

        all_objects = WritingImplement.objects.order_by('name')

        all_specializations = WritingImplement.specializations.order_by('name')

        eq_(len(all_objects), len(all_specializations))

        self._check_attributes(all_objects, all_specializations)

    def test_filter(self):
        """Check the filter() method works correctly"""

        filtered_objects = WritingImplement.objects \
            .filter(length__gte=10) \
            .filter(name__endswith='pen')

        filtered_specializations = WritingImplement.specializations \
            .filter(name__endswith='pen') \
            .filter(length__gte=10)

        self._check_attributes(filtered_objects, filtered_specializations)

        single_filter = WritingImplement.specializations.filter(
            name__endswith='pen', length__gte=10
            )

        eq_(single_filter[0], filtered_specializations[0])

    def test_exclude(self):
        """Check the exclude() method works correctly"""

        excluded_objects = WritingImplement.objects.exclude(length__lt=9)
        excluded_specializations = \
            WritingImplement.specializations.exclude(length__lt=9)

        self._check_attributes(excluded_objects, excluded_specializations)

    def test_slice_index(self):
        """
        Check that querysets can be sliced by a single index value correctly
        """

        all_objects = WritingImplement.objects.order_by('name')
        all_specializations = WritingImplement.specializations.order_by('name')

        eq_(len(all_objects), len(all_specializations))

        for i in xrange(len(all_objects)):
            o = all_objects[i]
            s = all_specializations[i]

            compare_generalization_to_specialization(o, s)

    def test_slice_range(self):
        """Test various range slices for compatibility"""

        # Two numbers:
        sliced_objects = WritingImplement.objects.order_by('name')[1:4]
        sliced_specializations = \
            WritingImplement.specializations.order_by('name')[1:4]

        self._check_attributes(sliced_objects, sliced_specializations)

        # Just end point:
        sliced_objects = WritingImplement.objects.order_by('length')[:3]
        sliced_specializations = \
            WritingImplement.specializations.order_by('length')[:3]

        self._check_attributes(sliced_objects, sliced_specializations)

        # Just start point:
        sliced_objects = WritingImplement.objects.order_by('-length')[1:]
        sliced_specializations = \
            WritingImplement.specializations.order_by('-length')[1:]

        self._check_attributes(sliced_objects, sliced_specializations)

    def test_order(self):
        """Test various orderings for compatibility"""

        # By name:
        ordered_objects = WritingImplement.objects.order_by('name')
        ordered_specializations = \
            WritingImplement.specializations.order_by('name')

        self._check_attributes(ordered_objects, ordered_specializations)

        # By inverse length and then name:
        ordered_objects = WritingImplement.objects.order_by('-length', 'name')
        ordered_specializations = WritingImplement.specializations.order_by(
            '-length', 'name'
            )

        self._check_attributes(ordered_objects, ordered_specializations)

    def test_get(self):
        """Check that the get() method behaves correctly"""

        general = WritingImplement.objects.get(name=PenData.GeneralPen.name)
        specialized = WritingImplement.specializations.get(
            name=PenData.GeneralPen.name
            )

        self._check_attributes([general], [specialized])

    def test_values(self):
        """Check values returns a ValuesQuerySet in both cases"""

        normal_values = WritingImplement.objects.values('pk', 'name')
        specialized_values = \
            WritingImplement.specializations.values('pk', 'name')

        ok_(isinstance(normal_values, ValuesQuerySet))
        ok_(isinstance(specialized_values, ValuesQuerySet))

        for normal_item, specialized_item in zip(
            normal_values, specialized_values
            ):

            eq_(normal_item['name'], specialized_item['name'])
            eq_(normal_item['pk'], specialized_item['pk'])

    def test_values_list(self):
        """Check values_list returns a ValuesListQuerySet in both cases"""

        normal_values = WritingImplement.objects.values_list('pk', 'length')
        specialized_values = WritingImplement.specializations.values_list(
            'pk', 'length'
            )

        ok_(isinstance(normal_values, ValuesListQuerySet))
        ok_(isinstance(specialized_values, ValuesListQuerySet))

        for (n_pk, n_length), (s_pk, s_length) in zip(
            normal_values, specialized_values
            ):

            eq_(n_pk, s_pk)
            eq_(n_length, s_length)

    def test_flat_values_list(self):
        """
        Check value_list with flat=True  returns a ValuesListQuerySet in both
        cases
        """

        normal_values = WritingImplement.objects.values_list('pk', flat=True)
        specialized_values = WritingImplement.specializations.values_list(
            'pk', flat=True
            )

        ok_(isinstance(normal_values, ValuesListQuerySet))
        ok_(isinstance(specialized_values, ValuesListQuerySet))

        eq_(list(normal_values), list(specialized_values))

    def test_aggregate(self):
        """Aggregations work on both types of querysets in the same manner"""

        normal_agg = WritingImplement.objects.aggregate(Avg('length'))

        specialized_agg = \
            WritingImplement.specializations.aggregate(Avg('length'))

        eq_(normal_agg[normal_agg.keys()[0]],
            specialized_agg[specialized_agg.keys()[0]]
            )

    def test_count(self):
        """Counts work over both types of querysets"""

        normal_count = WritingImplement.objects.filter(length__lt=13).count()
        specialized_count = \
            WritingImplement.objects.filter(length__lt=13).count()

        eq_(normal_count, specialized_count)

    def test_in_bulk(self):
        """In bulk works across both types of queryset"""

        ids = list(WritingImplement.objects.values_list('pk', flat=True))[2:]

        normal_bulk = WritingImplement.objects.in_bulk(ids)
        specialized_bulk = WritingImplement.specializations.in_bulk(ids)

        eq_(normal_bulk.keys(), specialized_bulk.keys())

        self._check_attributes(normal_bulk.values(), specialized_bulk.values())

    def test_update(self):
        """update() works the same across querysets"""

        original_lengths = list(
            WritingImplement.objects.order_by('length').values_list(
                'length', flat=True
                )
            )

        WritingImplement.specializations.all().update(length=1+F('length'))

        new_lengths = list(
            WritingImplement.objects.order_by('length').values_list(
                'length', flat=True
                )
            )

        for original_length, new_length in zip(original_lengths, new_lengths):
            eq_(original_length+1, new_length)

    def test_complex_query(self):
        """SpecializedQuerysets can be constructed from Q objects"""

        q_small = Q(length__lt=10)
        q_large = Q(length__gt=13)

        normal_objects = WritingImplement.objects.filter(q_small | q_large)
        specialized_objects = WritingImplement.specializations.filter(
            q_small | q_large
            )

        self._check_attributes(normal_objects, specialized_objects)




# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2011,2013, 2degrees Limited <2degrees-floss@googlegroups.com>.
# All Rights Reserved.
#
# This file is part of djeneralize <https://github.com/2degrees/djeneralize>,
# which is subject to the provisions of the BSD at
# <http://dev.2degreesnetwork.com/p/2degrees-license.html>. A copy of the
# license should accompany this distribution. THIS SOFTWARE IS PROVIDED "AS IS"
# AND ANY AND ALL EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST
# INFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
#
##############################################################################

from django.db import models

from djeneralize.models import BaseGeneralizationModel

__all__ = [
    'WritingImplement', 'Pencil', 'Pen', 'FountainPen', 'BallPointPen',
    'no_meta_factory', 'no_specialization_factory',
    'invalid_specialization_factory', 'abstract_specialization_factory',
    'base_generalization_with_specialization_factory'
    ]

#{ General model

class WritingImplement(BaseGeneralizationModel):
    name = models.CharField(max_length=30)
    length = models.IntegerField()

    def __unicode__(self):
        return self.name

#}

#{ Direct children of WritingImplement, i.e. first specializtion

class Pencil(WritingImplement):

    lead = models.CharField(max_length=2) # i.e. HB, B2, H5

    class Meta:
        specialization = 'pencil'

class Pen(WritingImplement):

    ink_colour = models.CharField(max_length=30)

    class Meta:
        specialization = 'pen'

#}

#{ Grand-children of WritingImplement, i.e. second degree of specialization

class FountainPen(Pen):

    nib_width = models.DecimalField(max_digits=3, decimal_places=2)

    class Meta:
        specialization = 'fountain_pen'


class BallPointPen(Pen):

    replaceable_insert = models.BooleanField(default=False)

    class Meta:
        specialization = 'ballpoint_pen'

#}

#{ Factories which are needed for testing:

def no_meta_factory():
    """Factory to mask TypeError in metaclass for testing"""

    class General(BaseGeneralizationModel):
        pass

    class Specialized(General):
        pass

    return Specialized


def no_specialization_factory():
    """Factory to mask TypeError in metaclass for testing"""

    class General(BaseGeneralizationModel):
        pass

    class Specialized(General):
        class Meta:
            verbose_name = 'my_model'

    return Specialized

def invalid_specialization_factory():
    """Factory to mask ValueError in metaclass for testing"""

    class General(BaseGeneralizationModel):
        pass

    class Specialized(General):
        class Meta:
            specialization = 'Naughty specialization!'

    return Specialized

def abstract_specialization_factory():
    """Factory to mask TypeError on incorrect specialization declaration"""

    class Abstract(BaseGeneralizationModel):
        class Meta:
            abstract = True
            specialization = 'abstract'

    return Abstract

def base_generalization_with_specialization_factory():
    """
    Factory to make a direct subclass of BaseGeneralizationModel which
    incorrectly declares specializations.

    """

    class General(BaseGeneralizationModel):

        class Meta:
            specialization = 'general'

    return General

#}


# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2011,2013, 2degrees Limited <2degrees-floss@googlegroups.com>.
# All Rights Reserved.
#
# This file is part of djeneralize <https://github.com/2degrees/djeneralize>,
# which is subject to the provisions of the BSD at
# <http://dev.2degreesnetwork.com/p/2degrees-license.html>. A copy of the
# license should accompany this distribution. THIS SOFTWARE IS PROVIDED "AS IS"
# AND ANY AND ALL EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT
# NOT LIMITED TO, THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST
# INFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
#
##############################################################################

from itertools import chain
import os
# Ensure that Django knows where the project is:
os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.test_djeneralize.settings'

from django.http import Http404
from fixture.django_testcase import FixtureTestCase
from nose.tools import (
    eq_, ok_, assert_false, raises, assert_raises, assert_not_equal
    )

from djeneralize.utils import *
from .fixtures import *
from .test_djeneralize.writing.models import *


class TestMetaclass(object):
    """Tests for the actions the metaclass performs"""


    def test_specializations_general(self):
        """
        The specializations dictionary on the most general case is populated
        with all the descdencts direct or otherwise.

        """

        specializations = WritingImplement._meta.specializations
        specializations_classes = set(specializations.values())
        specializations_keys = set(specializations.keys())

        ok_(Pen in specializations_classes)
        ok_(Pencil in specializations_classes)
        ok_(FountainPen in specializations_classes)
        ok_(BallPointPen in specializations_classes)

        ok_(Pen.model_specialization in specializations_keys)
        ok_(Pencil.model_specialization in specializations_keys)
        ok_(FountainPen.model_specialization in specializations_keys)
        ok_(BallPointPen.model_specialization in specializations_keys)


    def test_sub_specialization(self):
        """
        Only the child specializations are stored in a specialization of the
        general case.

        """

        specializations = Pen._meta.specializations
        specializations_classes = set(specializations.values())
        specializations_keys = set(specializations.keys())

        assert_false(Pen in specializations_classes)
        assert_false(Pencil in specializations_classes)
        ok_(FountainPen in specializations_classes)
        ok_(BallPointPen in specializations_classes)

        assert_false(Pen.model_specialization in specializations_keys)
        assert_false(Pencil.model_specialization in specializations_keys)
        ok_(FountainPen.model_specialization in specializations_keys)
        ok_(BallPointPen.model_specialization in specializations_keys)

        eq_(BallPointPen._meta.specializations, {})

    def test_path_specialization(self):
        """
        The specialization for each (sub-)class of the generalized model has it
        ancestry recorded.

        """

        eq_(WritingImplement.model_specialization, '/')
        eq_(Pen.model_specialization, '/pen/')
        eq_(Pencil.model_specialization, '/pencil/')
        eq_(FountainPen.model_specialization, '/pen/fountain_pen/')
        eq_(BallPointPen.model_specialization, '/pen/ballpoint_pen/')

    @raises(TypeError)
    def test_missing_meta(self):
        """All specializations must declare a inner class Meta"""

        no_meta_factory()

    @raises(TypeError)
    def test_missing_specialization(self):
        """
        All specializations must declare a inner class Meta which defines their
        specialization.

        """

        no_specialization_factory()

    @raises(ValueError)
    def test_invalid_specialization(self):
        """
        All declared specializations must contain only alphanumeric characters.

        """

        invalid_specialization_factory()

    @raises(TypeError)
    def test_abstract_specializations(self):
        """
        It is not permissible to declare specializations if the model is
        abstract.

        """

        abstract_specialization_factory()

    @raises(TypeError)
    def test_general_specializations(self):
        """
        It is not permissible to declare specializations if the model is
        a direct sub-class of BaseGeneralizationModel.

        """

        base_generalization_with_specialization_factory()


class TestFindNextPathDown(object):
    """Tests for find_next_path_down."""

    def test_root_down(self):
        """Test the path below the root path can be identified correctly."""

        root = '/'
        full_path = '/home/barry/dev/'

        eq_(find_next_path_down(root, full_path, '/'), '/home/')

    def test_non_root(self):
        """Test the path below a non-root path can be identified correctly"""

        non_root = '/home/'
        full_path = '/home/barry/dev/'

        eq_(find_next_path_down(non_root, full_path, '/'), '/home/barry/')


class TestModelInstance(FixtureTestCase):
    """Tests for model instances"""

    datasets = [PenData, PencilData, FountainPenData, BallPointPenData]

    def test_no_matching_specialization_raises_key_error(self):
        """
        If get_as_specialization is called on the most specialized instance a
        key error will be raised.

        """

        pencil = Pencil.objects.all()[0]

        assert_raises(KeyError, pencil.get_as_specialization)

    def test_final_specialization(self):
        """
        By default the get_as_specialization method returns the most specialized
        specialzation.

        """

        montblanc_general = WritingImplement.objects.get(name='Mont Blanc')
        montblanc_intermediate = Pen.objects.get(name='Mont Blanc')
        montblanc_special = FountainPen.objects.get(name='Mont Blanc')

        eq_(montblanc_general.get_as_specialization(), montblanc_special)

        eq_(montblanc_intermediate.get_as_specialization(), montblanc_special)

    def test_direct_specialization(self):
        """
        The get_as_specialization method can get the most direct specialization
        by setting final_specialization=False.

        """

        montblanc_general = WritingImplement.objects.get(name='Mont Blanc')
        montblanc_intermediate = Pen.objects.get(name='Mont Blanc')
        montblanc_special = FountainPen.objects.get(name='Mont Blanc')

        eq_(montblanc_general.get_as_specialization(False),
            montblanc_intermediate
            )

        eq_(montblanc_intermediate.get_as_specialization(False),
            montblanc_special
            )

    def test_default_specialization_type(self):
        """
        Ensure that the default specialization type is correctly set at
        creation time.

        """
        pencil = FountainPen()
        eq_(pencil.specialization_type, FountainPen.model_specialization)


class TestSpecializedQueryset(FixtureTestCase):
    """
    Tests for the specialized queryset (and therefore for the specialized
    manager as well).

    """

    datasets = [PenData, PencilData, FountainPenData, BallPointPenData]

    def setUp(self):
        # Transform the datasets into a dictionary keyed by name:
        datasets = {}

        all_datasets = chain(
            PenData.__dict__.items(), PencilData.__dict__.items(),
            FountainPenData.__dict__.items(), BallPointPenData.__dict__.items()
            )

        for class_name, inner_class in all_datasets:
            if class_name.startswith('_') or class_name == 'Meta':
                continue

            datasets[inner_class.name] = inner_class

        self.datasets = datasets

    def test_all_final(self):
        """
        The all() method on the manager and queryset returns final
        specializations.

        """

        all_writing_implements = WritingImplement.specializations.all()

        models = set(m.__class__ for m in all_writing_implements)

        assert_false(WritingImplement in models)
        ok_(Pen in models)
        ok_(Pencil in models)
        ok_(FountainPen in models)
        ok_(BallPointPen in models)


        # Ensure that all the objects have the correct fields and values
        # specified in their original definition, i.e. they've been
        # reconstituted correctly:
        for wi in all_writing_implements:
            dataset = self.datasets[wi.name]

            for field_name, value in dataset.__dict__.items():
                if field_name.startswith('_') or field_name == 'ref':
                    continue

                model_value = getattr(wi, field_name)
                eq_(model_value, value)

    def test_all_direct(self):
        """
        The all() method on the manager and queryset returns direct
        specializations for the specializations manager and calling the direct()
        method.

        """

        all_writing_implements = WritingImplement.specializations.direct().all()

        models = set(m.__class__ for m in all_writing_implements)

        assert_false(WritingImplement in models)
        ok_(Pen in models)
        ok_(Pencil in models)
        assert_false(FountainPen in models)
        assert_false(BallPointPen in models)

    def test_filter_chain_final(self):
        """
        Chained calls to the filter() method on the specialized queryset return
        final specializations

        """

        filtered_writing_implements = WritingImplement.specializations.filter(
            length__gt=10 # call to the manager
            ).filter(
            specialization_type__startswith='/pen/' # call to the queryset
            )

        models = set(m.__class__ for m in filtered_writing_implements)

        assert_false(WritingImplement in models)
        ok_(Pen in models)
        assert_false(Pencil in models)
        ok_(FountainPen in models)
        ok_(BallPointPen in models)

        expected_names = [
            'General pen', 'Mont Blanc', 'Parker', 'Bic', 'Papermate'
            ]

        for expected_name, wi in zip(expected_names, filtered_writing_implements):
            eq_(expected_name, wi.name)

    def test_filter_chain_direct(self):
        """
        Chained calls to the filter() method on the specialized queryset return
        direct specializations via the specializations manager and calling the
        direct()

        """

        filtered_writing_implements = WritingImplement.specializations.filter(
            length__gt=10 # call to the manager
            ).filter(
            specialization_type__startswith='/pen/' # call to the queryset
            ).direct()

        models = set(m.__class__ for m in filtered_writing_implements)

        assert_false(WritingImplement in models)
        ok_(Pen in models)
        assert_false(Pencil in models)
        assert_false(FountainPen in models)
        assert_false(BallPointPen in models)

        expected_names = [
            'General pen', 'Mont Blanc', 'Parker', 'Bic', 'Papermate'
            ]

        for expected_name, wi in zip(expected_names, filtered_writing_implements):
            eq_(expected_name, wi.name)

    def test_get_final(self):
        """
        Calling get() returns the final specialization when calling the
        specialization manager

        """

        mont_blanc = WritingImplement.specializations.get(name='Mont Blanc')

        eq_(mont_blanc.__class__, FountainPen)
        eq_(mont_blanc.nib_width, FountainPenData.MontBlanc.nib_width)
        eq_(mont_blanc.length, FountainPenData.MontBlanc.length)

        mont_blanc = Pen.specializations.get(name='Mont Blanc')

        eq_(mont_blanc.__class__, FountainPen)


    def test_get_direct(self):
        """
        Calling get() returns the direct specialization when calling the
        specialization manager and setting direct()

        """

        mont_blanc = WritingImplement.specializations.direct().get(
            name='Mont Blanc'
            )

        eq_(mont_blanc.__class__, Pen)

        mont_blanc = Pen.specializations.direct().get(name='Mont Blanc')

        eq_(mont_blanc.__class__, FountainPen)

    def test_final(self):
        """
        Calling the final() method on the manager or queryset ensures that the
        _final_specialization is set to True

        """

        qs = WritingImplement.specializations.final()
        ok_(qs._final_specialization)

        qs = WritingImplement.specializations.direct().final()
        ok_(qs._final_specialization)

    def test_direct(self):
        """
        Calling the direct() method on the manager or queryset ensures that the
        _final_specialization is set to False

        """

        qs = WritingImplement.specializations.direct()
        assert_false(qs._final_specialization)

        qs = WritingImplement.specializations.final().direct()
        assert_false(qs._final_specialization)

    def test_annotate(self):
        """Annotation is not possible for SpecializedQuerySets"""

        assert_raises(
            NotImplementedError, WritingImplement.specializations.annotate
            )

    def test_extra(self):
        """Queries added with .extra() are inherited in specializations."""
        writing_implement = WritingImplement.specializations.extra(
            select={'extra_field': 'SELECT 1'},
            )[0]

        ok_(hasattr(writing_implement, 'extra_field'))
        eq_(writing_implement.extra_field, 1)

    def test_ordering(self):
        """Ordering of the initial queryset is respected in the child objects"""

        lengths = sorted(
            WritingImplement.objects.values_list('length', flat=True)
            )

        ordered_wi = WritingImplement.specializations.order_by('length')

        for wi, expected_length in zip(ordered_wi, lengths):
            assert_not_equal(wi.__class__, WritingImplement)
            eq_(wi.length, expected_length)

    def test_slicing_range(self):
        """
        Slicing the queryset with a range gives specialized model instances.

        """

        ordered_wi = WritingImplement.specializations.order_by('length')[1:3]

        first_wi, second_wi = tuple(ordered_wi)

        eq_(first_wi, Pencil.objects.get(name='Technical'))
        eq_(second_wi, BallPointPen.objects.get(name='Bic'))

    def test_slicing_single_index(self):
        """
        Slicing the queryset with a single value gives a single specialized
        model instance.

        """

        wi = WritingImplement.specializations.order_by('length')[5]

        assert_not_equal(wi.__class__, WritingImplement)
        eq_(wi, Pen.objects.all()[0])

    def test_ordered_by_extra_field(self):
        """
        The ordering of a SpecializedQuerySet can be set by a field in an
        "extra" statement.

        """
        writing_implements = WritingImplement.specializations.extra(
            select={'extra_field': 'SELECT 1'},
            ).order_by('extra_field')
        eq_(writing_implements[0].extra_field, 1)

        reversed_writing_implements = WritingImplement.specializations.extra(
            select={'extra_field': 'SELECT 1'},
            ).order_by('-extra_field')

        eq_(reversed_writing_implements[0].extra_field, 1)

class TestGetSpecializationOr404(FixtureTestCase):
    """Tests for get_specialization_or_404"""

    datasets = [PenData, PencilData, FountainPenData, BallPointPenData]

    def test_model_class_exists(self):
        """
        get_specialization_or_404 returns the specialized model instance if it
        is called with a BaseSpecialized model class and parameters which
        correspond to a DB entry.

        """

        pencil = Pencil.objects.get(name=PencilData.Technical.name)
        eq_(pencil, get_specialization_or_404(
            WritingImplement, name=PencilData.Technical.name
            ))

    def test_model_class_does_not_exist(self):
        """
        get_specialization_or_404 raises a Http404 error if it is called with a
        BaseSpecialized model class and parameters which do not correspond to a
        DB entry.

        """

        assert_raises(
            Http404, get_specialization_or_404, WritingImplement,
            name='some thing else'
            )

    def test_manager_exists(self):
        """
        get_specialization_or_404 returns the specialized model instance if it
        is called with a SpecializedManager instance and parameters which
        correspond to a DB entry.

        """

        pencil = Pencil.objects.get(name=PencilData.Technical.name)
        eq_(pencil, get_specialization_or_404(
            WritingImplement.specializations, name=PencilData.Technical.name
            ))

    def test_manager_does_not_exist(self):
        """
        get_specialization_or_404 raises a Http404 error if it is called with a
        SpecializedManager instance and parameters which do not correspond to a
        DB entry.

        """

        assert_raises(
            Http404, get_specialization_or_404, WritingImplement.specializations,
            name='some thing else'
            )

    def test_queryset_exists(self):
        """
        get_specialization_or_404 returns the specialized model instance if it
        is called with a SpecializedQuerySet instance and parameters which
        correspond to a DB entry.

        """

        pencil = Pencil.objects.get(name=PencilData.Technical.name)
        eq_(pencil, get_specialization_or_404(
            WritingImplement.specializations.all(),
            name=PencilData.Technical.name
            ))

    def test_queryset_does_not_exist(self):
        """
        get_specialization_or_404 raises a Http404 error if it is called with a
        SpecializedQuerySet instance and parameters which do not correspond to a
        DB entry.

        """

        assert_raises(
            Http404, get_specialization_or_404,
            WritingImplement.specializations.all(), name='some thing else'
            )




"""Test suite to ensure that correct copying of managers"""

import os
# Ensure that Django knows where the project is:
os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.test_djeneralize.settings'

from django.db.models.manager import Manager
from fixture.django_testcase import FixtureTestCase
from nose.tools import (
    eq_, ok_, assert_false, raises, assert_raises, assert_not_equal
    )

from djeneralize.manager import SpecializationManager
from .test_djeneralize.fruit.models import *


class TestManager(object):

    def test_correct_manager_declared(self):
        """
        The specializations attribute of Fruit points to SpecializedFruitManager
        and not SpecializationManager.

        """

        ok_(isinstance(Fruit.specializations, SpecializedFruitManager),
            "Expected Fruit.specializations to be an instance of "
            "SpecializedFruitManager, but got %s" %
            Fruit.specializations.__class__
            )

    def test_default_manager(self):
        """The _default_manager isn't perturbed by djeneralize"""

        ok_(isinstance(Fruit._default_manager, FruitManager))

    def test_base_manager(self):
        """The _default_manager isn't perturbed by djeneralize"""

        ok_(isinstance(Fruit._base_manager, Manager))

    def test_default_specialization_manager_set(self):
        """
        The _default_specialization_manager should always be set on
        BaseGeneralizationModel sub-classes

        """

        ok_(hasattr(Fruit, '_default_specialization_manager'))
        ok_(isinstance(
            Fruit._default_specialization_manager, SpecializedFruitManager
            ), 'Expected Fruit._default_specialization_manager to be an '
            'instance of SpecializedFruitManager, but found %s' %
            Fruit._default_specialization_manager.__class__
            )

    def test_base_specialization_manager_set(self):
        """
        The _base_specialization_manager should always be set on
        BaseGeneralizationModel sub-classes

        """

        ok_(hasattr(Fruit, '_base_specialization_manager'))
        eq_(Fruit._base_specialization_manager.__class__, SpecializationManager)

    def test_specialized_not_inherit_specialized_managers(self):
        """
        Like normal managers, Specialized models do not copy the managers from
        their parent.

        """

        eq_(Apple.specializations.__class__, SpecializationManager)
        eq_(Banana.specializations.__class__, SpecializationManager)



from django.db import models

from djeneralize.fields import SpecializedForeignKey
from djeneralize.models import BaseGeneralizationModel


class Shop(models.Model):

    name = models.CharField(max_length=30)
    producer = SpecializedForeignKey('FruitProducer', related_name='shops')


class FruitProducer(BaseGeneralizationModel):

    name = models.CharField(max_length=30)
    pen = models.ForeignKey('writing.WritingImplement')
    produce = SpecializedForeignKey('fruit.Fruit')


class EcoProducer(FruitProducer):

    fertilizer = models.CharField(max_length=30)

    class Meta:
        specialization = 'eco_producer'


class StandardProducer(FruitProducer):

    chemicals = models.CharField(max_length=30)

    class Meta:
        specialization = 'standard_producer'
